@@ -1,15 +1,19 @@
+import boto3
 import json
 import logging
-import warnings
+import requests
 import traceback as tb
+import warnings
+from aws_requests_auth.boto_utils import BotoAWSRequestsAuth
+from copy import copy, deepcopy
+from daggerml._config import DML_API_ENDPOINT, DML_API_HOST, DML_REGION
 from dataclasses import dataclass
 from typing import NewType, Optional
-import daggerml._config as _conf
-from http.client import HTTPConnection, HTTPSConnection
-from urllib.parse import urlparse
 
 
 logger = logging.getLogger(__name__)
+conn_pool = requests.Session()
+boto3_session = boto3.session.Session()
 
 
 def _json_dumps_default(obj):
@@ -59,36 +63,28 @@ class NodeError(DmlError):
     pass
 
 
-def _api(api, op, group=_conf.DML_GROUP_ID, **kwargs):
+def _api(api, op, **kwargs):
     try:
-        url = urlparse(_conf.DML_API_ENDPOINT)
-        scheme = url.scheme
-        host = url.hostname
-        port = url.port or 443
-        path = url.path
-        assert all(x is not None for x in [scheme, host, port, path]), \
-            f'invalid endpoint URL: {_conf.DML_API_ENDPOINT}'
-        conn = (HTTPConnection if scheme == 'http' else HTTPSConnection)(host, port)
-        headers = {'content-type': 'application/json', 'accept': 'application/json'}
-        if _conf.DML_API_KEY is not None:
-            headers['x-daggerml-apikey'] = _conf.DML_API_KEY
-        if group is not None:
-            headers['x-daggerml-group'] = group
+        payload = dict(api=api, op=op, **kwargs)
         while True:
-            conn.request('POST', path, json.dumps(dict(api=api, op=op, **kwargs)), headers)
-            resp = conn.getresponse()
-            data = resp.read()
-            if resp.status != 504:
+            assert DML_API_ENDPOINT is not None, 'API endpoint not configured'
+            assert DML_API_HOST is not None and DML_REGION is not None, 'invalid API endpoint'
+            assert boto3_session.get_credentials() is not None, 'AWS credentials not found'
+            auth = BotoAWSRequestsAuth(aws_host=DML_API_HOST,
+                                       aws_region=DML_REGION,
+                                       aws_service='execute-api')
+            resp = conn_pool.post(DML_API_ENDPOINT, auth=auth, json=payload)
+            data = resp.json()
+            if resp.status_code != 504:
                 break
-        if resp.status != 200:
-            raise ApiError(f'{resp.status} {resp.reason}')
-        resp = json.loads(data)
-        if resp['status'] != 'ok':
-            err = resp['error']
+        if resp.status_code != 200:
+            raise ApiError(f'{resp.status_code} {resp.reason}')
+        if data['status'] != 'ok':
+            err = data['error']
             if err['context']:
-                logger.error('api error: %s', err['context'])
+                logger.error('api error: %s', err['message'] + '\n' + err['context'])
             raise ApiError(f'{err["code"]}: {err["message"]}')
-        return resp['result']
+        return data['result']
     except KeyboardInterrupt:
         raise
     except ApiError:
@@ -97,18 +93,6 @@ def _api(api, op, group=_conf.DML_GROUP_ID, **kwargs):
         raise ApiError(f'{e.__class__.__name__}: {str(e)}')
 
 
-def login(username, password):
-    """login to daggerml
-
-    Parameters
-    ----------
-    username : str
-    password : str
-    """
-    resp = _api('auth', 'create_api_key', username=username, password=password)
-    return resp
-
-
 def list_dags(name=None):
     """list all completed dags
 
@@ -149,7 +133,6 @@ def daggerml():
     from time import sleep
     from collections.abc import Mapping
     from weakref import WeakKeyDictionary
-    from uuid import uuid4
 
     tag2resource = {}
 
@@ -194,25 +177,15 @@ def daggerml():
         tag2resource[tag] = cls
         return
 
-    def to_data(py, dag=None):
+    def to_data(py):
         if isinstance(py, Node):
             return {'type': 'ref', 'value': {'node_id': py.id}}
-        if callable(py):
-            if dag is None:
-                raise RuntimeError(
-                    'cannot call `to_data` on a local function without a dag argument'
-                )
-            fn_id = py.__qualname__
-            if fn_id == '<lambda>':
-                fn_id += uuid4().hex
-            fn_id = py.__module__ + ':' + fn_id
-            py = [Node(dag, dag.executor_id), fn_id]
-        if isinstance(py, list) or isinstance(py, tuple):
-            return {'type': 'list', 'value': [to_data(x, dag) for x in py]}
-        elif isinstance(py, dict) or isinstance(py, Mapping):
+        elif isinstance(py, (list, tuple)):
+            return {'type': 'list', 'value': [to_data(x) for x in py]}
+        elif isinstance(py, (dict, Mapping)):
             if not all([isinstance(x, str) for x in py]):
                 raise TypeError('map datum keys must be strings')
-            return {'type': 'map', 'value': {k: to_data(v, dag) for (k, v) in py.items()}}
+            return {'type': 'map', 'value': {k: to_data(v) for (k, v) in py.items()}}
         elif isinstance(py, type(None)):
             return {'type': 'scalar', 'value': {'type': 'null'}}
         elif isinstance(py, str):
@@ -263,48 +236,71 @@ def daggerml():
         id: str
 
         def __len__(self):
-            return len(self.dag.to_py(self))
+            _meta = self.meta['$dml']
+            if 'length' in _meta:
+                return _meta['length']
+            keys = _meta.get('keys')
+            if keys is not None:
+                return len(keys)
+            raise ValueError('cannot iterate of this node type')
 
-        def __getitem__(self, key):
-            f = Node(self.dag, self.dag.get_fn)
-            resp = f(self, key)
-            cached = CACHE.get(self)
-            if cached is not None:  # populate value in cache from cache (local funcs)
-                if isinstance(key, Node):
-                    key = key.to_py()
-                CACHE[resp] = cached[key]
+        def __iter__(self):
+            _meta = self.meta['$dml']
+            if _meta['type'] == 'list':
+                for i in range(_meta['length']):
+                    yield self[i]
+            elif _meta['type'] == 'map':
+                for key in _meta['keys']:
+                    yield key
+            else:
+                raise ValueError('cannot iterate of this node type')
+
+        def __add__(self, other):
+            if len(other) == 0:
+                return self
+            f = self.dag.from_py([self.dag.db_executor, 'concat', self, other])
+            resp = f()
             return resp
 
-        def __call__(self, *args, block=True, meta={}):
-            args = [self.dag.from_py(x) for x in args]
-            expr = [self.id] + [x.id for x in args]
-            if callable(CACHE.get(self)):
-                resp = _api('dag', 'put_fnapp_and_claim', dag_id=self.dag.id,
-                            ttl=0, expr=expr, meta=meta)
-                if resp['success']:
-                    return Node(self.dag, resp['node_id'])
-                if resp['error'] is not None:
-                    logger.debug('ignoring error: %s', json.dumps(resp['error']))
-                try:
-                    result = CACHE[self](*args)
-                    resp2 = _api('node', 'commit_node',
-                                 node_id=resp['node_id'],
-                                 secret=self.dag.secret,
-                                 token=resp['refresh_token'],
-                                 data=to_data(result, dag=self.dag))
-                    assert resp2['finalized'], 'failed to finalize node'
-                except Exception as e:
-                    err = format_exception(e)
-                    _api('node', 'fail_node', secret=self.dag.secret,
-                         node_id=resp['node_id'], token=resp['refresh_token'],
-                         error=err)
-                    raise NodeError(err)
-                n = Node(self.dag, resp['node_id'])
-                CACHE[n] = result
-                return n
-            waiter = NodeWaiter(self.dag, expr, meta)
-            if not block:
-                return waiter
+        def __getitem__(self, key):
+            f = self.dag.from_py([self.dag.db_executor, 'get', self, key])
+            resp = f()
+            if self in CACHE:
+                if isinstance(key, Node):
+                    key = key.to_py()
+                CACHE[resp] = CACHE[self][key]
+            return resp
+
+        def call_async(self, *args, **meta):
+            """call a remote function asynchronously
+
+            Parameters
+            ----------
+            *args : dml types
+            **meta : json serializable
+
+            Returns
+            -------
+            NodeWaiter
+            """
+            expr = self + args
+            waiter = NodeWaiter(self.dag, expr.id, meta)
+            return waiter
+
+        def __call__(self, *args, **meta):
+            """call a remote function
+
+            Parameters
+            ----------
+            *args : dml types
+            **meta : json serializable
+
+            Returns
+            -------
+            Node
+            """
+            expr = self + args
+            waiter = NodeWaiter(self.dag, expr.id, meta)
             waiter.wait(2)
             return waiter.result
 
@@ -334,27 +330,26 @@ def daggerml():
             self._result = None
             self.check()
 
+        def __hash__(self):  # required for this to be a key in a map
+            return hash(self.id)
+
         def check(self):
-            self._resp = _api(
-                'dag',
-                'put_fnapp',
-                dag_id=self.dag.id,
-                expr=self.expr,
-                meta=self.meta,
-                secret=self.dag.secret
-            )
+            self._resp = _api('dag', 'put_fnapp',
+                              dag_id=self.dag.id,
+                              expr=self.expr,
+                              meta=self.meta,
+                              secret=self.dag.secret)
             return self.result
 
         @property
-        def node_id(self):
+        def id(self):
             return self._resp['node_id']
 
         @property
         def result(self):
             if self._resp['success']:
-                return Node(self.dag, self.node_id)
+                return Node(self.dag, self.id)
             if self._resp['error'] is not None:
-                logger.debug('ignoring error: %s', json.dumps(self._resp['error']))
                 raise NodeError(self._resp['error'])
             return
 
@@ -369,21 +364,21 @@ def daggerml():
         id: str
         name: str = None
         version: int = None
-        group: str = None
         expr_id: str = None
-        get_fn: str = None
+        db_exec: str = None
+        # get_fn: str = None
         executor_id: str = None
         secret: str = None
 
         @classmethod
-        def new(cls, name, version=None, group=_conf.DML_GROUP_ID):
+        def new(cls, name, version=None):
             """create a new dag"""
-            resp = _api('dag', 'create_dag', name=name, version=version, group=group)
+            resp = _api('dag', 'create_dag', name=name, version=version)
             if resp is not None:
-                return cls(**resp, group=group)
+                return cls(**resp)
 
         @classmethod
-        def from_claim(cls, executor, secret, ttl, group, node_id=None):
+        def from_claim(cls, executor, secret, ttl, node_id=None):
             """claim a remote execution
 
             Parameters
@@ -394,18 +389,19 @@ def daggerml():
                 the executor's secret
             ttl : int
                 how long for the claim to be active for (before needing to refresh)
-            group : str
-                the group ID
             node_id : str, optional
                 the specific node_id to claim
             """
             if isinstance(executor, Resource):
                 executor = executor.to_dict()
-            resp = _api('node', 'claim_node', executor=executor,
-                        ttl=ttl, node_id=node_id, group=group, secret=secret)
+            resp = _api('node', 'claim_node',
+                        executor=executor,
+                        ttl=ttl,
+                        node_id=node_id,
+                        secret=secret)
             if resp is None:
                 return
-            return cls(**resp, group=group)
+            return cls(**resp)
 
         @property
         def expr(self):
@@ -417,53 +413,69 @@ def daggerml():
             """this dag's executor resource"""
             return Node(self, self.executor_id).to_py()
 
+        @property
+        def db_executor(self):
+            return Node(self, self.db_exec).to_py()
+
         def from_py(self, py):
             """convert a python datastructure to a literal node"""
-            if isinstance(py, Node):
-                return py
-            res = _api('dag', 'put_literal', dag_id=self.id, data=to_data(py, self),
-                       group=self.group, secret=self.secret)
+            res = _api('dag', 'put_literal',
+                       dag_id=self.id,
+                       data=to_data(py),
+                       secret=self.secret)
             node = Node(self, res['node_id'])
             if node not in CACHE:
-                CACHE[node] = py
+                CACHE[node] = deepcopy(py)
             return node
 
         def to_py(self, node):
-            """convert a node to a python datastructure"""
-            if node.dag != self:
-                raise ValueError('node does not belong to dag')
-            if node in CACHE:
-                return CACHE[node]
-            py = from_data(_api('node', 'get_node', node_id=node.id,
-                                group=self.group, secret=self.secret))
-            CACHE[node] = py
-            return py
+            """convert a [collection of] nodes to a python datastructure"""
+            if isinstance(node, Node):
+                if node.dag != self:
+                    raise ValueError('node does not belong to dag')
+                if node in CACHE:
+                    py = self.to_py(CACHE[node])
+                else:
+                    py = from_data(_api('node', 'get_node',
+                                        node_id=node.id,
+                                        secret=self.secret))
+                    CACHE[node] = deepcopy(py)
+            elif isinstance(node, (tuple, list)):
+                py = tuple([self.to_py(x) for x in node])
+            elif isinstance(node, Mapping):
+                py = {k: self.to_py(v) for k, v in node.items()}
+            elif isinstance(node, (bool, str, int, float, type(None), Resource)):
+                py = node
+            return copy(py)
 
         def fail(self, failure_info={}):
             """fail a dag"""
             if isinstance(failure_info, (dict, list, tuple, Resource)):
                 failure_info = json.loads(json_dumps(failure_info))
-            _api('dag', 'fail_dag', dag_id=self.id, group=self.group,
-                 secret=self.secret, failure_info=failure_info)
+            _api('dag', 'fail_dag',
+                 dag_id=self.id,
+                 secret=self.secret,
+                 failure_info=failure_info)
             return
 
         def commit(self, result):
             """commit a dag result"""
             result = self.from_py(result)
-            _api('dag', 'commit_dag', dag_id=self.id, result=result.id,
-                 group=self.group, secret=self.secret)
+            _api('dag', 'commit_dag',
+                 dag_id=self.id,
+                 result=result.id,
+                 secret=self.secret)
             return
 
         def delete(self):
             """delete a dag (must be failed / committed first)"""
-            _api('dag', 'delete_dag', dag_id=self.id,
-                 group=self.group, secret=self.secret)
+            _api('dag', 'delete_dag', dag_id=self.id, secret=self.secret)
             return
 
         def refresh(self, ttl=300):
             """refresh a dag claim (ttl is same as in `from_claim`)"""
             res = _api('dag', 'refresh_claim', dag_id=self.id,
-                       group=self.group, secret=self.secret, ttl=ttl,
+                       secret=self.secret, ttl=ttl,
                        refresh_token=self.id)
             if res is None:
                 raise DagError('Failed to refresh dag!')
@@ -481,7 +493,7 @@ def daggerml():
         def create_resource(self, tag=None):
             """get a resource object"""
             res = _api('dag', 'create_resource', dag_id=self.id,
-                       group=self.group, secret=self.secret, tag=tag)
+                       secret=self.secret, tag=tag)
             return Node(self, res['node_id']), res['secret']
 
         def __repr__(self):
